from .external_evaluation import HansenSengupta, Krawczyk
# from .optimize import globopt, Uni, Tol, outliers
from .optimize import globopt, Uni, outliers
from .solution_existence import miranda

from .recognizing_functional import Tol



from .RealInterval import ClassicalArithmetic, KaucherArithmetic, ArrayInterval, precision, SingleInterval, Interval, ARITHMETICS, INTERVAL_CLASSES, single_type
from .utils import *
from .visualization import lineqs, IntLinIncR2, lineqs3D, IntLinIncR3, OneShotVisual2D, scatter_plot

from .linear import *
from .nonlinear import *



from bisect import bisect_left
import numpy as np

from intvalpy.RealInterval import Interval
from intvalpy.utils import asinterval


class KeyWrapper:
    def __init__(self, iterable, key):
        self.it = iterable
        self.key = key

    def __getitem__(self, i):
        return self.key(self.it[i])

    def __len__(self):
        return len(self.it)


def globopt(func, x0, tol=1e-12, maxiter=2000):
    Y = x0.copy
    y = func(Y).a
    L = [(Y, y)]

    nit = 0
    while func(Y).wid >= tol and nit <= maxiter:
        l = np.argmax(Y.wid)
        Y1 = L[0][0].copy
        Y2 = L[0][0].copy
        Y1[l], Y2[l] = Interval(Y[l].a, Y[l].mid, sortQ=False), Interval(Y[l].mid, Y[l].b, sortQ=False)

        v1, v2 = func(Y1).a, func(Y2).a
        del L[0]

        newcol = (Y1, v1)
        bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
        L.insert(bslindex, newcol)

        newcol = (Y2, v2)
        bslindex = bisect_left(KeyWrapper(L, key=lambda c: c[1]), newcol[1])
        L.insert(bslindex, newcol)
        Y = L[0][0]
        nit += 1

    return L[0][0], func(L[0][0])


def __tolsolvty(func, grad, a, b, weight=None, x0=None,
                tol_f=1e-12, tol_x=1e-12, tol_g=1e-12, maxiter=2000):

    nsims = 30
    alpha = 2.3
    hs = 1
    nh = 3
    q1 = 0.9
    q2 = 1.1

    if grad is None:
        raise Exception('Необходимо задать вектор градиент!')

    m = len(a)
    n = len(grad)
    if weight is None:
        weight = np.ones(m)

    bc = b.mid
    br = b.rad
    tt = None

    def calcfg(x):
        index = x >= 0
        infsup = bc - func(a, x)

        tt = weight * (br - np.maximum(np.abs(infsup.a), np.abs(infsup.b)))
        mc = np.argmin(tt)
        if n == 1:
            _grad = asinterval(grad[0](a[mc], x))
        else:
            _grad = asinterval([g(a[mc], x) for g in grad])

        if -infsup[mc].a <= infsup[mc].b:
            dd = weight[mc] * (_grad.a * index + _grad.b * (~index))
        else:
            dd = -weight[mc] * (_grad.b * index + _grad.a * (~index))

        return tt[mc], dd, tt


    if x0 is None:
        x = np.zeros(n)
    else:
        x = x0

    B = np.eye(n)
    vf = np.zeros(nsims) + float('inf')
    w = 1./alpha - 1

    f, g0, _ = calcfg(x)
    ff = f;  xx = x
    cal = 1;  ncals = 1
    ccode = False

    for nit in range(int(maxiter)):
        vf[nsims-1] = ff
        if np.linalg.norm(g0) < tol_g:
            ccode = True
            break

        g1 = B.T @ g0
        g = B @ (g1/np.linalg.norm(g1))
        normg = np.linalg.norm(g)

        r = 1
        cal = 0
        deltax = 0

        while r > 0 and cal <= 500:
            cal += 1
            x = x + hs * g
            deltax = deltax + hs * normg
            f, g1, tt = calcfg(x)

            if f > ff:
                ff = f
                xx = x

            if np.mod(cal, nh) == 0:
                hs = hs * q2
            r = g @ g1

        if cal > 500:
            ccode = False
            break

        if cal == 1:
            hs = hs * q1

        ncals = ncals + cal
        if deltax < tol_x:
            ccode = True
            break

        dg = B.T @ (g1 - g0)
        xi = dg / np.linalg.norm(dg)
        B = B + w * np.outer((B @ xi), xi)
        g0 = g1

        vf = np.roll(vf, 1)
        vf[0] = abs(ff - vf[0])

        if abs(ff) > 1:
            deltaf = np.sum(vf)/abs(ff)
        else:
            deltaf = np.sum(vf)

        if deltaf < tol_f:
            ccode = True
            break
        ccode = False

    return ccode, xx, ff, tt


def Uni(func, a, b, x=None, maxQ=False, grad=None, weight=None, x0=None, tol=1e-12, maxiter=2000):

    br = b.rad
    bm = b.mid

    def __uni(x):
        data = func(a, x)
        return np.min(br - (bm - data).mig)
    __minus_uni = lambda x: -__uni(x)

    if maxQ:
        if x0 is None:
            _, x0, _, _ = __tolsolvty(func, grad, a, b, weight=weight, x0=None, maxiter=maxiter,
                               tol_f=1e-8, tol_x=1e-8, tol_g=1e-8)

        from scipy.optimize import minimize
        maximize = minimize(__minus_uni, x0, method='Nelder-Mead', tol=tol,
                            options={'maxiter': maxiter})
        return maximize.success, maximize.x, -maximize.fun

    else:
        if x is None:
            x = np.zeros(len(a))
        return __uni(x)



def _tol_tsopt(model, a, b, grad, weight=None, x0=None, tol=1e-12, maxiter=2000):
    return __tolsolvty(model, grad, a, b, weight=weight, x0=x0, maxiter=maxiter,
                       tol_f=tol, tol_x=tol, tol_g=tol)[:-1]


def _tol_iopt(model, a, b, grad, x0, tol=1e-12, maxiter=2000):
    return None, None, None


def Tol(func, a, b, x=None, maxQ=False, grad=None, weight=None, x0=None, tol=1e-12, maxiter=2000):

    if maxQ:
        return __tolsolvty(func, grad, a, b, weight=weight, x0=x0, maxiter=maxiter,
                           tol_f=tol, tol_x=tol, tol_g=tol)[:-1]

    else:
        br = b.rad
        bm = b.mid

        def __tol(x):
            data = func(a, x)
            return np.min(br - (bm - data).mag)

        if x is None:
            x = np.zeros(len(a))

        return __tol(x)


def outliers(func, a, b, grad, functional='tol', weight=None, x0=None, tol=1e-12, maxiter=2000, method='standard deviations'):

    def interquartile(data):
        q25, q75 = np.percentile(data, 25), np.percentile(data, 75)
        iqr = q75 - q25
        cut_off = iqr * 1.5

        lower, upper = q25 - cut_off, q75 + cut_off
        return np.argwhere((data < lower) | (data > upper)).flatten()

    def standard_deviations(data):
        # Set upper and lower limit to 3 standard deviation
        std, mean = np.std(data), np.mean(data)
        cut_off = std * 3

        lower, upper = mean - cut_off, mean + cut_off
        return np.argwhere((data < lower) | (data > upper)).flatten()

    WorkLista = asinterval(a).copy
    WorkListb = asinterval(b).copy

    if functional == 'tol':
        _, _, _, tt = __tolsolvty(func, grad, a, b, weight=weight, x0=x0, maxiter=maxiter,
                                  tol_f=tol, tol_x=tol, tol_g=tol)
    else:
        Exception('Данный функционал не предусмотрен.')

    if method == 'standard deviations':
        outliers_index = standard_deviations(tt)
    elif method == 'interquartile':
        outliers_index = interquartile(tt)
    else:
        Exception('Данный метод не предусмотрен.')

    index = np.delete(np.arange(WorkLista.shape[0]), outliers_index)
    WorkLista = WorkLista[index]
    WorkListb = WorkListb[index]

    return WorkLista, WorkListb, outliers_index, tt



from intvalpy.utils import zeros


def miranda(func, x):
    n = len(x)

    def l(i):
        result = zeros(n)
        for k in range(n):
            if k == i:
                result[k] = x[k].a
            else:
                result[k] = x[k]
        return result

    def u(i):
        result = zeros(n)
        for k in range(n):
            if k == i:
                result[k] = x[k].b
            else:
                result[k] = x[k]
        return result

    for k in range(n):
        if (func(l(k))[k].b <= 0 and func(u(k))[k].a >= 0) or \
           (func(l(k))[k].a >= 0 and func(u(k))[k].b <= 0):
            pass
        else:
            return False
    return True



import io
import os
from setuptools import setup, find_packages

os.environ['CVXOPT_BUILD_GLPK'] = '1'

readme_file = os.path.join(os.path.dirname(__file__), 'README.md')
with io.open(readme_file, mode='r', encoding='utf-8') as f:
    README = f.read()

INSTALL_REQUIRES = [
    'matplotlib',
    'numpy',
    'scipy',
    'cvxopt',
    'mpmath'
]

setup(
    name='intvalpy',
    version='1.5.5',
    description='An interval library in Python that uses classical interval ' + \
                'arithmetic and Kauсher arithmetic + Kahan division in some functions',
    long_description=README,
    long_description_content_type='text/markdown',
    license='MIT License',
    keywords=[
        'Interval',
        'inequality visualization',
        'optimal solutions',
        'math',
        'range'
    ],
    author='Андросов Артем Станиславович, Шарый Сергей Петрович',
    author_email='artem.androsov@gmail.com, shary@ict.nsc.ru',
    url='https://github.com/AndrosovAS/intvalpy',
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES
)



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from .RealInterval import INTERVAL_CLASSES

from .utils import infinity


def scatter_plot(x, y, title="Box", color='gray', alpha=0.5, s=10, size=(15, 15), save=False):
    """
    A mathematical diagram depicting the values of two variables in the form of bars on a Cartesian plane.

    Parameters:

        x: Interval
            Interval vector of the data position on the OX axis.

        y: Interval
            Interval vector of the data position on the OY axis.

        title: str, optional
            The top legend of the graph.

        color: str, optional
            The color of the display of the bars.

        alpha: float, optional
            Transparency of the graph.

        s: float, optional
            How big are the points of the vertices.

        size: tuple, optional
            The size of the drawing window.

        save: bool, optional
            If the value is True, the graph is saved.

    Returns:

        out: None
            A scatter plot is displayed.
    """

    ox = np.array([x.a, x.a, x.b, x.b])
    oy = np.array([y.a, y.b, y.b, y.a])

    fig = plt.figure(figsize=size)
    ax = fig.add_subplot(111, title=title)

    plt.fill(ox, oy, color=color, alpha=alpha)
    index = ((x.a - x.b) == 0) & ((y.a - y.b) == 0)
    ax.scatter(ox[:, index], oy[:, index], s=s, color=color, alpha=alpha)

    if save:
        fig.savefig(title + ".png")


def Unique(a, decimals=12):
    a = np.ascontiguousarray(a)
    a = np.around(a, decimals=int(decimals))
    _, index = np.unique(a.view([('', a.dtype)]*a.shape[1]), return_index=True)
    index = sorted(index)

    return a[index]


def non_repeat(a, b):
    a = np.copy(np.ascontiguousarray(a))
    a = np.around(a, decimals=15)
    a1 = (a.T - b).T
    _, index = np.unique(a1.view([('', a1.dtype)]*a1.shape[1]), return_index=True)
    index = sorted(index)

    return a[index], b[index]


def clear_zero_rows(a, b, ndim=2):
    a, b = np.ascontiguousarray(a), np.ascontiguousarray(b)
    a, b = np.around(a, decimals=15), np.around(b, decimals=15)

    cnmty = True
    if np.sum((np.sum(abs(a) <= 1e-15, axis=1) == ndim) & (b > 0)) > 0:
        cnmty = False

    index = np.where(np.sum(abs(a) <= 1e-15, axis=1) != ndim)
    return a[index], b[index], cnmty


def BoundaryIntervals(A, b):
    m, n = A.shape
    S = []

    for i in range(m):
        q = [-infinity, infinity]
        si = True
        dotx = (A[i]*b[i])/np.dot(A[i], A[i])

        p = np.array([-A[i, 1], A[i, 0]])

        for k in range(m):
            if k == i:
                continue
            Akx = np.dot(A[k], dotx)
            c = np.dot(A[k], p)

            if np.sign(c) == -1:
                tmp = (b[k] - Akx) / c
                q[1] = q[1] if q[1] <= tmp else tmp
            elif np.sign(c) == 1:
                tmp = (b[k] - Akx) / c
                q[0] = q[0] if tmp < q[0] else tmp
            else:
                if Akx < b[k]:
                    if np.dot(A[k], A[i]) > 0:
                        si = False
                        break
                    else:
                        return []

        if q[0] > q[1]:
            si = False

        # избавление от неопределённости inf * 0
        p = p + 1e-301
        if si:
            S.append(list(dotx+p*q[0]) + list(dotx+p*q[1]) + [i])

    return np.array(S)


def ParticularPoints(S, A, b):
    PP = []
    V = S[:, :2]

    binf = ~((abs(V[:, 0]) < float("inf")) & (abs(V[:, 1]) < float("inf")))

    if len(V[binf]) != 0:
        nV = 1
        for k in S[:, 4]:
            k = int(k)
            PP.append((A[k]*b[k])/np.dot(A[k], A[k]))
    else:
        nV = 0
        PP = V

    return np.array(PP), nV, binf


def Intervals2Path(S):
    bs, bp = S[0, :2], S[0, :2]
    P = [bp]

    while len(S) > 0:
        for k in range(len(S)):
            if np.max(np.abs(bs - S[k, :2])) < 1e-8:
                index = k
                break
        es = S[index, 2:4]

        if np.max(np.abs(bs-es)) > 1e-8:
            P.append(es)

            if np.max(np.abs(bp-es)) < 1e-8:
                return np.array(P)
            bs = es
        S = np.delete(S, index, axis=0)
    return np.array(P)


def ChangeVariable(A, b, k):
    y1 = np.zeros(3)
    xi = (A[k] * b[k]) / (A[k] @ A[k])

    l = np.argmin(np.abs(A[k]))
    l1 = ((l + 1) % 3)
    l2 = ((l1 + 1) % 3)

    y1[l1], y1[l2] = A[k, l2], -A[k, l1]
    y2 = np.cross(A[k], y1)

    index = np.array([l for l in range(len(A)) if l != k])
    A, b = A[index], b[index]

    At = np.array([A @ y1, A @ y2]).T
    bt = b - A @ xi

    return xi, At, bt, y1, y2


__center_rm = []
def lineqs(A, b, show=True, title="Solution Set", color='gray',
           bounds=None, alpha=0.5, s=10, size=(15, 15), save=False):
    """
    The function visualizes the set of solutions of a system of linear algebraic
    inequalities A x >= b with two variables by the method of boundary intervals, and
    also outputs the vertices of the set of solutions.

    If the set of solutions is unlimited, then the algorithm independently
    selects the rendering boundaries.

    Parameters:

            A: float, array_like
                Matrix of a system of linear algebraic inequalities.

            b: float, array_like
                The vector of the right part of the system of linear algebraic inequalities.

            show: bool, optional
                This parameter is responsible for whether a set of solutions will be shown.
                By default, the value is set to True, i.e. the graph is being drawn.

            title: str, optional
                The top legend of the graph.

            color: str, optional
                The color of the inner area of the set of solutions.

            bounds: array_like, optional
                Borders of the drawing area. The first element of the array is responsible for the lower faces
                on the OX and OY axes, and the second for the upper ones. Thus, in order to OX
                lay within [-2, 2], and OY within [-3, 4], it is necessary to set bounds as [[-2, -3], [2, 4]].

            alpha: float, optional
                Transparency of the graph.

            s: float, optional
                How big are the points of the vertices.

            size: tuple, optional
                The size of the drawing window.

            save: bool, optional
                If the value is True, the graph is saved.

    Returns:

            out: list
                Returns a list of ordered vertices.
                If show = True, then the graph is drawn.
    """

    if isinstance(A, INTERVAL_CLASSES) or isinstance(b, INTERVAL_CLASSES):
        raise Exception('Interval type system!')

    A = np.asarray(A)
    b = np.asarray(b)

    n, m = A.shape
    assert m <= 2, "There should be two columns in matrix A."
    assert b.shape[0] == n, "The size of the matrix A must be consistent with the size of the vector of the right part of b."

    A, b, cnmty = clear_zero_rows(A, b)

    S = BoundaryIntervals(A, b)
    if len(S) == 0:
        return S

    PP, nV, binf = ParticularPoints(S, A, b)

    if (np.asarray([binf]) == True).any():
        if bounds is None:
            PP = np.array(PP)
            PPmin, PPmax = np.min(PP, axis=0), np.max(PP, axis=0)
            center = (PPmin + PPmax)/2
            rm = max((PPmax - PPmin)/2)
            __center_rm.append([max(abs(center) + 5*rm)])
            A = np.append(np.append(A, np.eye(2)), -np.eye(2)).reshape((len(A)+4, 2))
            b = np.append(np.append(b, center-5*rm), -(center+5*rm))

        else:
            A = np.append(np.append(A, np.eye(2)), -np.eye(2)).reshape((len(A)+4, 2))
            b = np.append(np.append(b, [bounds[0][0], bounds[0][1]]),
                          [-bounds[1][0], -bounds[1][1]])

        S = BoundaryIntervals(A, b)

    vertices = Intervals2Path(S)

    if show:
        fig = plt.figure(figsize=size)
        ax = fig.add_subplot(111, title=title)

        x, y = vertices[:, 0], vertices[:, 1]
        ax.fill(x, y, linestyle='-', linewidth=1, color=color, alpha=alpha)
        ax.scatter(x, y, s=s, color='black', alpha=1)

        if save:
            fig.savefig(title + ".png")
    return Unique(vertices)


def OneShotVisual2D(*args, title="Solution Set", grid=True, size=(15, 15), labelsize=None, save=False):

    fig = plt.figure(figsize=size)
    ax = fig.add_subplot(111, title=title)

    if grid:
        ax.grid()

    if labelsize is None:
        ax.xaxis.set_tick_params(labelsize=size[0])
        ax.yaxis.set_tick_params(labelsize=size[0])
    else:
        ax.xaxis.set_tick_params(labelsize=labelsize)
        ax.yaxis.set_tick_params(labelsize=labelsize)

    for v in args:
        alpha = 0.5
        s = 10
        color = 'gray'
        vertices = None
        for key in v.keys():
            if key == "vertices":
                vertices = v["vertices"]
            elif key == "alpha":
                alpha = v["alpha"]
            elif key == "s":
                s = v["s"]
            elif key == "color":
                color = v["color"]

        if not (vertices is None):
            if isinstance(vertices, list):
                for k in range(4):
                    if len(vertices[k]) > 0:
                        x, y = vertices[k][:, 0], vertices[k][:, 1]
                        ax.fill(x, y, linestyle='-', linewidth=1, color=color, alpha=alpha)
                        ax.scatter(x, y, s=s, color='black', alpha=1)
            else:
                if len(vertices) > 0:
                    x, y = vertices[:, 0], vertices[:, 1]
                    ax.fill(x, y, linestyle='-', linewidth=1, color=color, alpha=alpha)
                    ax.scatter(x, y, s=s, color='black', alpha=1)

    if save:
        fig.savefig(title + ".png")


def IntLinIncR2(A, b, show=True, title="Solution Set", consistency='uni',
                bounds=None, color='gray', alpha=0.5, s=10, size=(15, 15), save=False):
    """
    The function visualizes a set of solutions of an interval system of linear
    algebraic equations A x = b with two variables by the boundary value method
    intervals, and also outputs the vertices of the set of solutions.

    If the set of solutions is unlimited, then the algorithm independently
    selects the rendering boundaries.

    Parameters:

            A: Interval
                The input interval matrix of ISLAE, which can be either square or rectangular.

            b: Interval
                The interval vector of the right part of the ISLAE.

            show: bool, optional
                This parameter is responsible for whether a set of solutions will be shown.
                By default, the value is set to True, i.e. the graph is being drawn.

            title: str, optional
                The top legend of the graph.

            consistency: str, optional
                A parameter for selecting the type of a set of solutions.
                If consistency = "uni", then the function returns the united set of the solution.
                If consistency = "tol", then the tolerance set of the solution.

            bounds: array_like, optional
                Borders of the drawing area. The first element of the array is responsible for the lower faces
                on the OX and OY axes, and the second for the upper ones. Thus, in order to OX
                lay within [-2, 2], and OY within [-3, 4], it is necessary to set bounds as [[-2, -3], [2, 4]].

            color: str, optional
                The color of the inner area of the set of solutions.

            alpha: float, optional
                Transparency of the graph.

            s: float, optional
                How big are the points of the vertices.

            size: tuple, optional
                The size of the drawing window.

            save: bool, optional
                If the value is True, the graph is saved.

    Returns:

            out: list
                Returns a list of ordered vertices in each ortant starting from the first
                and making a detour in a positive direction. If show = True, then the graph is drawn.
    """

    if not (isinstance(A, INTERVAL_CLASSES) or isinstance(b, INTERVAL_CLASSES)):
        return lineqs(A, b, show=show, title=title, color=color, bounds=bounds,
                      alpha=alpha, s=s, size=size, save=save)

    ortant = [(1, 1), (-1, 1), (-1, -1), (1, -1)]
    vertices = []
    n, m = A.shape

    assert m <= 2, "There should be two columns in matrix A."
    assert b.shape[0] == n, "The size of the matrix A must be consistent with the size of the vector of the right part of b."

    def algo(bounds):
        for ort in range(4):
            tmp = A.copy
            WorkListA = np.zeros((2*n+m, m))
            WorkListb = np.zeros(2*n+m)

            for k in range(m):
                if ortant[ort][k] == -1:
                    tmp[:, k] = tmp[:, k].dual
                WorkListA[2*n+k, k] = -ortant[ort][k]

            if consistency == 'uni':
                WorkListA[:n], WorkListA[n:2*n] = tmp.a, -tmp.b
                WorkListb[:n], WorkListb[n:2*n] = b.b, -b.a
            elif consistency == 'tol':
                WorkListA[:n], WorkListA[n:2*n] = -tmp.a, tmp.b
                WorkListb[:n], WorkListb[n:2*n] = -b.a, b.b
            else:
                msg = "Неверно указан тип согласования системы! Используйте 'uni' или 'tol'."
                raise Exception(msg)

            vertices.append(lineqs(-WorkListA, -WorkListb, show=False, bounds=bounds))
    algo(bounds)

    # Если в каком-либо ортанте множество решений неограничено, то создаём
    # новое отрисовочное окно, чтобы срез решений был одинаковым.
    global __center_rm
    if len(__center_rm) > 0:
        vertices = []
        _max = max(np.array(__center_rm))
        bounds = np.array([[-_max, -_max], [_max, _max]])
        algo(bounds)
    __center_rm = []

    if show:
        fig = plt.figure(figsize=size)
        ax = fig.add_subplot(111, title=title)

        for k in range(4):
            if len(vertices[k]) > 0:
                x, y = vertices[k][:, 0], vertices[k][:, 1]
                ax.fill(x, y, linestyle='-', linewidth=1, color=color, alpha=alpha)
                ax.scatter(x, y, s=s, color='black', alpha=1)
        if save:
            fig.savefig(title + ".png")

    return vertices


def lineqs3D(A, b, show=True, color='C0', alpha=0.5, s=10, size=(8, 8),
             bounds=None):
    """
    The function visualizes the set of solutions of a system of linear algebraic
    inequalities A x >= b with three variables by the method of boundary intervals, and
    also outputs the vertices of the set of solutions.

    If the set of solutions is unlimited, then the algorithm independently
    selects the rendering boundaries.


    Parameters:

        A: float, array_like
            Matrix of a system of linear algebraic inequalities.

        b: float, array_like
            The vector of the right part of the system of linear algebraic inequalities.

        show: bool, optional
            This parameter is responsible for whether a set of solutions will be shown.
            By default, the value is set to True, i.e. the graph is being drawn.

        color: str, optional
            The color of the inner area of the set of solutions.

        bounds: array_like, optional
            Borders of the drawing area. The first element of the array is responsible
            for the lower faces along the axes OX, OY and OZ, and the second for the upper ones.
            Thus, in order for OX to lie within [-2, 2], OY within [-3, 4], and OZ within [1, 5],
            it is necessary to set bounds as [[-2, -3, 1], [2, 4, 5]].

        alpha: float, optional
            Transparency of the graph.

        s: float, optional
            How big are the points of the vertices.

        size: tuple, optional
            The size of the drawing window.

    Returns:

        out: list
            Returns a list of ordered vertices.
            If show = True, then the graph is drawn.
    """

    A = np.asarray(A)
    b = np.asarray(b)

    A, b, cnmty = clear_zero_rows(A, b, ndim=3)
    A, b = non_repeat(A, b)

    n, m = A.shape
    assert m == 3, "There should be three columns in matrix A."
    assert b.shape[0] == n, "The size of the matrix A must be consistent with the size of the vector of the right part of b."

    V = []
    PP = []
    cfinite = True

    for k in range(n):
        xi, At, bt, y1, y2 = ChangeVariable(A, b, k)
        At, bt, cnmtyt = clear_zero_rows(At, bt)

        if not cnmtyt:
            continue

        if len(bt) == 0:
            PP.append(xi)
            cfinite = False
            continue

        St = BoundaryIntervals(At, bt)
        if len(St) == 0:
            continue

        PPt, nV, binf = ParticularPoints(St, At, bt)

        PPtt = []
        for l in range(len(PPt)):
            PPtt.append((xi + y1*PPt[l, 0]) + y2*PPt[l, 1])

        PP.append(PPtt)

        if (np.asarray([binf]) == True).any():
            cfinite = False
        if not nV:
            V.append(PPtt)

    if len(PP) == 0:
        return []

    if not cfinite:
        if bounds is None:
            x, y, z = [], [], []
            for pp in PP:
                for el in pp:
                    x.append(el[0])
                    y.append(el[1])
                    z.append(el[2])

            x = np.array(x)
            y = np.array(y)
            z = np.array(z)

            xmin, ymin, zmin = np.min(x), np.min(y), np.min(z)
            xmax, ymax, zmax = np.max(x), np.max(y), np.max(z)

            center = np.array([0, 0, 0]) + (x + y + z).mean()
            rm = (xmax-xmin) + (ymax-ymin) + (zmax-zmin)

            if rm <= 1e-14:
                rm = rm + 1
            __center_rm.append([max(abs(center) + 1*rm)])
            A = np.append(np.append(A, np.eye(3)), -np.eye(3)).reshape((len(A)+6, 3))
            b = np.append(np.append(b, center-rm), -(center+rm))
        else:
            A = np.append(np.append(A, np.eye(3)), -np.eye(3)).reshape((len(A)+6, 3))
            b = np.append(np.append(b, [bounds[0][0], bounds[0][1], bounds[0][2]]),
                          [-bounds[1][0], -bounds[1][1], -bounds[1][2]])

    vertices = []
    mn = len(b)
    for i in range(mn):
        xi, At, bt, y1, y2 = ChangeVariable(A, b, i)
        At, bt, cnmtyt = clear_zero_rows(At, bt)

        if not cnmtyt:
            continue

        St = BoundaryIntervals(At, bt)
        if len(St) > 0:
            Pt = Intervals2Path(St)
            P = []
            for l in range(len(Pt)):
                P.append((xi + y1*Pt[l, 0]) + y2*Pt[l, 1])

            P = np.array(P)
            vertices.append(P)

    if show:
        fig = plt.figure(figsize=size)
        ax = fig.add_subplot(111, projection='3d')

        if not (bounds is None):
            ax.set_xlim((bounds[0][0], bounds[1][0]))
            ax.set_ylim((bounds[0][1], bounds[1][1]))
            ax.set_zlim((bounds[0][2], bounds[1][2]))

        l = 0
        for v in vertices:
            if l >= n:
                color = 'red'

            x, y, z = v[:, 0], v[:, 1], v[:, 2]

            poly3d = [list(zip(x, y, z))]
            PC = Poly3DCollection(poly3d, linewidths=1)
            PC.set_alpha(alpha)
            PC.set_facecolor(color)
            ax.add_collection3d(PC)

            ax.plot(x, y, z, color='black', alpha=1)
            ax.scatter(x, y, z, s=s, color='black')
            l += 1

    return vertices


def IntLinIncR3(A, b, show=True, consistency='uni', color='C0',
                alpha=0.5, s=10, size=(8, 8), bounds=None, zero_lvl=True):
    """
    The function visualizes a set of solutions of an interval system of linear
    algebraic equations A x = b with three variables by the boundary value method
    intervals, and also outputs the vertices of the set of solutions.

    If the set of solutions is unlimited, then the algorithm independently
    selects the rendering boundaries.


    Parameters:

        A: Interval
            The input interval matrix of ISLAE, which can be either square or rectangular.

        b: Interval
            The interval vector of the right part of the ISLAE.

        show: bool, optional
            This parameter is responsible for whether a set of solutions will be shown.
            By default, the value is set to True, i.e. the graph is being drawn.

        consistency: str, optional
            A parameter for selecting the type of a set of solutions.
            If consistency = "uni", then the function returns the united set of the solution.
            If consistency = "tol", then the tolerance set of the solution.

        bounds: array_like, optional
            Borders of the drawing area. The first element of the array is responsible
            for the lower faces along the axes OX, OY and OZ, and the second for the upper ones.
            Thus, in order for OX to lie within [-2, 2], OY within [-3, 4], and OZ within [1, 5],
            it is necessary to set bounds as [[-2, -3, 1], [2, 4, 5]].

        color: str, optional
            The color of the inner area of the set of solutions.

        alpha: float, optional
            Transparency of the graph.

        s: float, optional
            How big are the points of the vertices.

        size: tuple, optional
            The size of the drawing window.

    Returns:

        out: list
            Returns a list of ordered vertices in each ortant starting from the first
            and making a detour in a positive direction. If show = True, then the graph is drawn.
    """

    if not (isinstance(A, INTERVAL_CLASSES) or isinstance(b, INTERVAL_CLASSES)):
        return lineqs3D(A, b, show=False, color=color,
                        alpha=alpha, s=s, size=size, bounds=bounds)

    ortant = [(1, 1, 1), (-1, 1, 1), (-1, -1, 1), (1, -1, 1),
              (1, 1, -1), (-1, 1, -1), (-1, -1, -1), (1, -1, -1)]
    vertices = []
    n, m = A.shape

    assert m <= 3, "There should be three columns in matrix A."
    assert b.shape[0] == n, "The size of the matrix A must be consistent with the size of the vector of the right part of b."

    def algo(bounds):
        gxmin, gymin, gzmin = infinity, infinity, infinity
        gxmax, gymax, gzmax = -infinity, -infinity, -infinity
        for ort in range(8):
            tmp = A.copy
            WorkListA = np.zeros((2*n+m, m))
            WorkListb = np.zeros(2*n+m)

            for k in range(m):
                if ortant[ort][k] == -1:
                    tmp[:, k] = tmp[:, k].dual
                WorkListA[2*n+k, k] = -ortant[ort][k]

            if consistency == 'uni':
                WorkListA[:n], WorkListA[n:2*n] = tmp.a, -tmp.b
                WorkListb[:n], WorkListb[n:2*n] = b.b, -b.a
            elif consistency == 'tol':
                WorkListA[:n], WorkListA[n:2*n] = -tmp.a, tmp.b
                WorkListb[:n], WorkListb[n:2*n] = -b.a, b.b
            else:
                msg = "Неверно указан тип согласования системы! Используйте 'uni' или 'tol'."
                raise Exception(msg)

            for k in range(2*n):
                if WorkListb[k] >= 0:
                    WorkListb[k] -= 1e-15
                else:
                    WorkListb[k] += 1e-15

            vertices.append(lineqs3D(-WorkListA, -WorkListb, show=False, bounds=bounds))
            for v in vertices[-1]:
                gxmin = min(gxmin, np.min(v[:, 0]))
                gymin = min(gymin, np.min(v[:, 1]))
                gzmin = min(gzmin, np.min(v[:, 2]))

                gxmax = max(gxmax, np.max(v[:, 0]))
                gymax = max(gymax, np.max(v[:, 1]))
                gzmax = max(gzmax, np.max(v[:, 2]))
        return gxmin, gymin, gzmin, gxmax, gymax, gzmax

    gxmin, gymin, gzmin, gxmax, gymax, gzmax = algo(bounds)

    # Если в каком-либо ортанте множество решений неограничено, то создаём
    # новое отрисовочное окно, чтобы срез решений был одинаковым.
    global __center_rm
    if len(__center_rm) > 0:
        vertices = []
        _max = max(np.array(__center_rm))
        bounds = np.array([[-_max, -_max, -_max], [_max, _max, _max]])
        gxmin, gymin, gzmin, gxmax, gymax, gzmax = algo(bounds)
    __center_rm = []

    if show:
        fig = plt.figure(figsize=size)
        ax = fig.add_subplot(111, projection='3d')

        if not bounds is None:
            ax.set_xlim((bounds[0][0], bounds[1][0]))
            ax.set_ylim((bounds[0][1], bounds[1][1]))
            ax.set_zlim((bounds[0][2], bounds[1][2]))

        color1 = color
        for el in vertices:

            color = color1
            l = 0
            for v in el:
                x, y, z = v[:, 0], v[:, 1], v[:, 2]
                xmin, ymin, zmin = np.min(abs(x)), np.min(abs(y)), np.min(abs(z))
                xmax, ymax, zmax = np.max(abs(x)), np.max(abs(y)), np.max(abs(z))

                if n <= l and l < n + 3 and (xmin * ymin * zmin == 0):
                    l += 1
                    continue

                elif zero_lvl and ((xmax == xmin and xmax == 0 and abs(gxmin*gxmax) > 1e-14) or
                                   (ymax == ymin and ymax == 0 and abs(gymin*gymax) > 1e-14) or
                                   (zmax == zmin and zmax == 0 and abs(gzmin*gzmax) > 1e-14)):      # demo
                    continue

                elif l >= n:
                    color = 'red'

                poly3d = [list(zip(x, y, z))]
                PC = Poly3DCollection(poly3d, linewidths=1)
                PC.set_alpha(alpha)
                PC.set_facecolor(color)
                ax.add_collection3d(PC)

                ax.plot(x, y, z, color='black', alpha=1)
                ax.scatter(x, y, z, s=s, color='black')
                l += 1

    return vertices
